import equinox as eqx
import jax.numpy as jnp
from equinox import filter_jit
from jax import vmap
from jax.lax import cond

from .utils import format_jax_array


class AbstractKernel(eqx.Module):
	@filter_jit
	def __call__(self, x1, x2=None):
		# If no x2 is provided, we compute the covariance between x1 and itself
		if x2 is None:
			x2 = x1

		# Turn scalar inputs into vectors
		x1, x2 = jnp.atleast_1d(x1), jnp.atleast_1d(x2)

		# Call the appropriate method
		if jnp.ndim(x1) == 1 and jnp.ndim(x2) == 1:
			return self.static_class.pairwise_cov_if_not_nan(self, x1, x2)
		elif jnp.ndim(x1) == 2 and jnp.ndim(x2) == 1:
			return self.static_class.cross_cov_vector_if_not_nan(self, x1, x2)
		elif jnp.ndim(x1) == 1 and jnp.ndim(x2) == 2:
			return self.static_class.cross_cov_vector_if_not_nan(self, x2, x1)
		elif jnp.ndim(x1) == 2 and jnp.ndim(x2) == 2:
			return self.static_class.cross_cov_matrix(self, x1, x2)
		else:
			raise ValueError(
				f"Invalid input dimensions: x1 has shape {x1.shape}, x2 has shape {x2.shape}. "
				"Expected scalar, 1D or 2D arrays as inputs."
			)

	def __add__(self, other):
		from kernax.OperatorKernels import SumKernel

		return SumKernel(self, other)

	def __radd__(self, other):
		from kernax.OperatorKernels import SumKernel

		return SumKernel(other, self)

	def __sub__(self, other):
		from kernax.OperatorKernels import SumKernel
		from kernax.WrapperKernels import NegKernel

		return SumKernel(self, NegKernel(other))

	def __rsub__(self, other):
		from kernax.OperatorKernels import SumKernel
		from kernax.WrapperKernels import NegKernel

		return SumKernel(other, NegKernel(self))

	def __neg__(self):
		from kernax.WrapperKernels import NegKernel

		return NegKernel(self)

	def __mul__(self, other):
		from kernax.OperatorKernels import ProductKernel

		return ProductKernel(self, other)

	def __rmul__(self, other):
		from kernax.OperatorKernels import ProductKernel

		return ProductKernel(other, self)

	def __str__(self):
		# FIXME: do not call `format_jax_array` when the pytree is filled with non-float values
		#  For example, try to print the `batch_in_axes` property of a BatchKernel
		# Print parameters, aka elements of __dict__ that are jax arrays
		return f"{self.__class__.__name__}({', '.join([f'{key}={format_jax_array(value)}' for key, value in self.__dict__.items() \
		if isinstance(value, jnp.ndarray)])})"


class StaticAbstractKernel:
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern: AbstractKernel, x1: jnp.ndarray, x2: jnp.ndarray) -> jnp.ndarray:
		"""
		Compute the kernel covariance value between two vectors.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		return jnp.array(jnp.nan)  # To be overwritten in subclasses

	@classmethod
	@filter_jit
	def pairwise_cov_if_not_nan(
		cls, kern: AbstractKernel, x1: jnp.ndarray, x2: jnp.ndarray
	) -> jnp.ndarray:
		"""
		Returns NaN if either x1 or x2 is NaN, otherwise calls the compute_scalar method.

		:param kern: kernel instance containing the hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		return cond(
			jnp.any(jnp.isnan(x1) | jnp.isnan(x2)),
			lambda _: jnp.nan,
			lambda _: cls.pairwise_cov(kern, x1, x2),
			None,
		)

	@classmethod
	@filter_jit
	def cross_cov_vector(
		cls, kern: AbstractKernel, x1: jnp.ndarray, x2: jnp.ndarray
	) -> jnp.ndarray:
		"""
		Compute the kernel cross covariance values between an array of vectors (matrix) and a vector.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: scalar array
		:return: vector array (N, )
		"""
		return vmap(lambda x: cls.pairwise_cov_if_not_nan(kern, x, x2), in_axes=0)(x1)

	@classmethod
	@filter_jit
	def cross_cov_vector_if_not_nan(
		cls, kern: AbstractKernel, x1: jnp.ndarray, x2: jnp.ndarray, **kwargs
	) -> jnp.ndarray:
		"""
		Returns an array of NaN if scalar is NaN, otherwise calls the compute_vector method.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: scalar array
		:param kwargs: hyperparameters of the kernel
		:return: vector array (N, )
		"""
		return cond(
			jnp.any(jnp.isnan(x2)),
			lambda _: jnp.full(len(x1), jnp.nan),
			lambda _: cls.cross_cov_vector(kern, x1, x2),
			None,
		)

	@classmethod
	@filter_jit
	def cross_cov_matrix(
		cls, kern: AbstractKernel, x1: jnp.ndarray, x2: jnp.ndarray
	) -> jnp.ndarray:
		"""
		Compute the kernel covariance matrix between two vector arrays.

		:param kern: kernel instance containing the hyperparameters
		:param x1: vector array (N, )
		:param x2: vector array (M, )
		:return: matrix array (N, M)
		"""
		return vmap(lambda x: cls.cross_cov_vector_if_not_nan(kern, x2, x), in_axes=0)(x1)



import equinox as eqx
import jax.numpy as jnp
from equinox import filter_jit
from jax import Array

from kernax import AbstractKernel, StaticAbstractKernel
from .utils import format_jax_array


class StaticConstantKernel(StaticAbstractKernel):
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern, x1: jnp.ndarray, x2: jnp.ndarray) -> jnp.ndarray:
		"""
		Compute the kernel covariance value between two vectors.

		:param kern: the kernel to use, containing hyperparameters
		:param x1: scalar array
		:param x2: scalar array
		:return: scalar array
		"""
		return kern.value  # The constant value is returned regardless of the inputs


class ConstantKernel(AbstractKernel):
	value: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticConstantKernel

	def __init__(self, value=1.0):
		"""
		Instantiates a constant kernel with the given value.

		:param value: the value of the constant kernel
		"""
		super().__init__()
		self.value = value

	def __str__(self):
		return format_jax_array(self.value)




import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from kernax import AbstractKernel, StaticAbstractKernel


class StaticPeriodicKernel(StaticAbstractKernel):
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern, x1: jnp.ndarray, x2: jnp.ndarray) -> jnp.ndarray:
		"""
		Compute the periodic kernel covariance value between two vectors.

				:param kern: the kernel to use, containing hyperparameters (length_scale, variance, period).
				:param x1: scalar array
				:param x2: scalar array
				:return: covariance value (scalar)
		"""
		dist = jnp.linalg.norm(x1 - x2)

		return kern.variance * jnp.exp(
			-2 * jnp.sin(jnp.pi * dist / kern.period) ** 2 / kern.length_scale**2
		)


class PeriodicKernel(AbstractKernel):
	length_scale: Array = eqx.field(converter=jnp.asarray)
	variance: Array = eqx.field(converter=jnp.asarray)
	period: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticPeriodicKernel

	def __init__(self, length_scale, variance, period):
		"""
		:param length_scale: length scale parameter (ℓ)
		:param variance: variance parameter (σ²)
		:param period: period parameter (p)
		"""
		super().__init__()
		self.length_scale = length_scale
		self.variance = variance
		self.period = period



import equinox as eqx
from equinox import filter_jit
from jax import Array
from jax import numpy as jnp

from kernax import AbstractKernel, StaticAbstractKernel


class StaticRationalQuadraticKernel(StaticAbstractKernel):
	@classmethod
	@filter_jit
	def pairwise_cov(cls, kern, x1: jnp.ndarray, x2: jnp.ndarray) -> jnp.ndarray:
		"""
		Compute the Rational Quadratic kernel covariance value between two vectors.

		:param kern: the kernel to use, containing hyperparameters (variance, length_scale, alpha)
		:param x1: scalar array
		:param x2: scalar array
		:return: covariance value (scalar)
		"""
		squared_dist = jnp.sum((x1 - x2) ** 2)

		base = 1 + squared_dist / (2 * kern.alpha * kern.length_scale**2)

		return kern.variance * jnp.power(base, -kern.alpha)


class RationalQuadraticKernel(AbstractKernel):
	length_scale: Array = eqx.field(converter=jnp.asarray)
	variance: Array = eqx.field(converter=jnp.asarray)
	alpha: Array = eqx.field(converter=jnp.asarray)
	static_class = StaticRationalQuadraticKernel

	def __init__(self, length_scale, variance, alpha):
		"""
		:param length_scale: length scale parameter (ℓ)
		:param variance: variance (σ²)
		:param alpha: relative weighting of large-scale and small-scale variations (α)
		"""
		super().__init__()
		self.length_scale = length_scale
		self.variance = variance
		self.alpha = alpha



"""
kernax-ml: A JAX-based kernel library for Gaussian Processes.

kernax-ml provides a collection of kernel functions (covariance functions) for
Gaussian Process models, with support for automatic differentiation, JIT
compilation, and composable kernel operations.
"""

__version__ = "0.1.8-alpha"
__author__ = "S. Lejoly"
__email__ = "simon.lejoly@unamur.be"
__license__ = "MIT"

from .AbstractKernel import AbstractKernel, StaticAbstractKernel
from .ConstantKernel import ConstantKernel, StaticConstantKernel
from .LinearKernel import LinearKernel, StaticLinearKernel
from .MaternKernels import (
	Matern12Kernel,
	Matern32Kernel,
	Matern52Kernel,
	StaticMatern12Kernel,
	StaticMatern32Kernel,
	StaticMatern52Kernel,
)
from .OperatorKernels import OperatorKernel, ProductKernel, SumKernel
from .PeriodicKernel import PeriodicKernel, StaticPeriodicKernel
from .RationalQuadraticKernel import RationalQuadraticKernel, StaticRationalQuadraticKernel
from .SEKernel import SEKernel, StaticSEKernel
from .WrapperKernels import (
	ActiveDimsKernel,
	ARDKernel,
	BatchKernel,
	DiagKernel,
	ExpKernel,
	LogKernel,
	NegKernel,
	WrapperKernel,
	BlockKernel,
	BlockDiagKernel,
)

__all__ = [
	# Package metadata
	"__version__",
	"__author__",
	"__email__",
	"__license__",
	# Base classes
	"StaticAbstractKernel",
	"AbstractKernel",
	# Base kernels
	"StaticSEKernel",
	"SEKernel",
	"StaticConstantKernel",
	"ConstantKernel",
	"StaticLinearKernel",
	"LinearKernel",
	"StaticPeriodicKernel",
	"PeriodicKernel",
	"StaticRationalQuadraticKernel",
	"RationalQuadraticKernel",
	# Matern family
	"StaticMatern12Kernel",
	"Matern12Kernel",
	"StaticMatern32Kernel",
	"Matern32Kernel",
	"StaticMatern52Kernel",
	"Matern52Kernel",
	# Composite kernels
	"OperatorKernel",
	"SumKernel",
	"ProductKernel",
	# Wrapper kernels
	"WrapperKernel",
	"NegKernel",
	"ExpKernel",
	"LogKernel",
	"DiagKernel",
	"BatchKernel",
	"ActiveDimsKernel",
	"ARDKernel",
	"BlockKernel",
	"BlockDiagKernel",
]
